@@ -31,8 +31,9 @@ import Xdp from 'gi://Xdp';
 const Format = imports.format;
 
 import {Application} from './application.js';
+import * as Gfx from './gfx.js';
 import {Overpass} from './overpass.js';
-import { lookupType } from './osmTypes.js';
+import { getTypeNameForPlace } from './osmTypes.js';
 import {Place} from './place.js';
 import * as PlaceIcons from './placeIcons.js';
 import {PlaceViewImage} from './placeViewImage.js';
@@ -69,8 +70,10 @@ export class PlaceView extends Gtk.Box {
                                                    'thumbnail-separator',
                                                    'label-title',
                                                    'label-icon',
+                                                   'secondary-label-icon',
                                                    'native-name',
                                                    'source-icon',
+                                                   'secondary-source-icon',
                                                    'source-label',
                                                    'source-box',
                                                    'address-label',
@@ -81,8 +84,10 @@ export class PlaceView extends Gtk.Box {
                                                    'title-box' ]);
         this._title = ui.labelTitle;
         this._icon = ui.labelIcon;
+        this._secondaryIcon = ui.secondaryLabelIcon;
         this._nativeName = ui.nativeName;
         this._sourceIcon = ui.sourceIcon;
+        this._secondarySourceIcon = ui.secondarySourceIcon;
         this._sourceLabel = ui.sourceLabel;
         this._sourceBox = ui.sourceBox;
         this._thumbnail = ui.bubbleThumbnail;
@@ -124,7 +129,7 @@ export class PlaceView extends Gtk.Box {
                                            hexpand: true });
         this.content.attach(this._placeDetails, 0, 0, 1, 1);
 
-        if (this.place.isCurrentLocation || !this.place.osmType) {
+        if (this.place.isRawCoordinates) {
             this._populate(this.place);
         } else {
             let overpass = new Overpass();
@@ -133,7 +138,8 @@ export class PlaceView extends Gtk.Box {
             if (placeItem !== null) {
 
                 // If the place is stale, update from Overpass.
-                if (placeItem.isStale()) {
+                if (placeItem.isStale() &&
+                    this.place.osmType !== GeocodeGlib.PlaceOsmType.UNKNOWN) {
                     overpass.populatePlace(this.place,
                                            this._onOverpass.bind(this));
                 } else {
@@ -146,9 +152,11 @@ export class PlaceView extends Gtk.Box {
                  */
                 if (this.place.prefilled) {
                     this._onOverpass(true);
-                } else {
+                } else if (this.place.osmType !== GeocodeGlib.PlaceOsmType.UNKNOWN) {
                     overpass.populatePlace(this.place,
                                            this._onOverpass.bind(this));
+                } else {
+                    this._populate(this.place);
                 }
             } else {
                 this._populate(this.place);
@@ -191,6 +199,19 @@ export class PlaceView extends Gtk.Box {
         this._mainStack.set_visible_child(val ? this._spinner : this._mainBox);
     }
 
+    /* get a suitable icon pixel size given a paintable and a maximum desired
+     * pixel height
+     */
+    _calculatePixelSizeForShield(paintable, pixelHeight) {
+        const intrinsicWidth = paintable.get_intrinsic_width();
+        const intrinsicHeight = paintable.get_intrinsic_height();
+
+        return intrinsicHeight < pixelHeight ?
+               Math.max(intrinsicWidth, intrinsicHeight) :
+               Math.max((intrinsicWidth / intrinsicHeight) * pixelHeight,
+                        pixelHeight);
+    }
+
     updatePlaceDetails() {
         let place = this.place;
         let formatter = new PlaceFormatter(place);
@@ -209,19 +230,47 @@ export class PlaceView extends Gtk.Box {
         }
 
         const title = formatter.title;
-        const typeName = lookupType(place.osmKey, place.osmValue);
+        const typeName = getTypeNameForPlace(place);
+        const shieldPaintables =
+            Gfx.drawShieldsForPlace(place, 2, this.get_scale_factor());
 
         /* if the place has a title, show it, otherwise if there's a known
          * place type name, show that together with a place type icon
          */
-        if (title) {
+        if (place.isRawCoordinates &&
+            place.name === place.coordinatesDescription) {
+            this._title.visible = true;
+            this._title.label = _("Coordinates");
+            this._icon.visible = false;
+            this._seondaryIcon.visible = false;
+        } else if (title) {
+            this._title.visible = true;
             this._title.label = formatter.title;
             this._icon.visible = false;
+            this._secondaryIcon.visible = false;
         } else if (typeName) {
             this._title.label = typeName;
-            this._icon.icon_name = PlaceIcons.getIconForPlace(place);
+            if (shieldPaintables.length > 0) {
+                this._icon.paintable = shieldPaintables[0];
+                this._icon.pixel_size =
+                    this._calculatePixelSizeForShield(shieldPaintables[0], 24);
+                if (shieldPaintables.length > 1) {
+                    this._secondaryIcon.paintable = shieldPaintables[1];
+                    this._secondaryIcon.pixel_size =
+                        this._calculatePixelSizeForShield(shieldPaintables[1], 24);
+                    this._secondaryIcon.visible = true;
+                } else {
+                    this._secondaryIcon.visible = false;
+                }
+            } else {
+                this._icon.icon_name = PlaceIcons.getIconForPlace(place);
+                this._icon.pixel_size = -1;
+                this._secondaryIcon.visible = false;
+            }
+            this._title.visible = true;
             this._icon.visible = true;
         } else {
+            this._title.visible = false;
             this._icon.visible = false;
         }
 
@@ -244,6 +293,15 @@ export class PlaceView extends Gtk.Box {
              */
             if (this._nativeName.get_layout().get_unknown_glyphs_count() === 0)
                 this._nativeName.visible = true;
+        } else if (place.name === place.nativeName && place.hiraganaName &&
+                   place.name !== place.hiraganaName) {
+            /* if the displayed name and the native name are identical, show
+             * a Japanese Hiragana name form when available providing a
+             * furigana-like pronounciation guide when displaying Japanese
+             * native name in a Japanese locale.
+             */
+            this._nativeName.label = place.hiraganaName;
+            this._nativeName.visible = true;
         }
 
         /* show the source (for shapelayer points), or if there has a known type
@@ -255,8 +313,33 @@ export class PlaceView extends Gtk.Box {
             this._sourceLabel.label = place.source;
             this._sourceBox.visible = true;
         } else if (title && typeName && place.osmKey !== 'place') {
+            if (shieldPaintables.length > 0) {
+                this._sourceIcon.get_style_context().remove_class('dim-label');
+                this._sourceIcon.paintable = shieldPaintables[0];
+                this._sourceIcon.pixel_size =
+                    this._calculatePixelSizeForShield(shieldPaintables[0], 18);
+
+                if (shieldPaintables.length > 1) {
+                    this._secondarySourceIcon.paintable = shieldPaintables[1];
+                    this._secondarySourceIcon.pixel_size =
+                        this._calculatePixelSizeForShield(shieldPaintables[1], 18);
+                    this._secondarySourceIcon.visible = true;
+                } else {
+                    this._secondarySourceIcon.visible = false;
+                }
+            } else {
+                const iconName = PlaceIcons.getIconForPlace(place);
+
+                // don't dim the icon for non-symbolic (transit network) icons
+                if (iconName.endsWith('-symbolic'))
+                    this._sourceIcon.get_style_context().add_class('dim-label');
+                else
+                    this._sourceIcon.get_style_context().remove_class('dim-label');
+                this._sourceIcon.icon_name = iconName;
+                this._sourceIcon.pixel_size = -1;
+                this._secondarySourceIcon.visible = false;
+            }
             this._sourceLabel.label = typeName;
-            this._sourceIcon.icon_name = PlaceIcons.getIconForPlace(place);
             this._sourceBox.visible = true;
         } else {
             this._sourceBox.visible = false;
@@ -264,9 +347,6 @@ export class PlaceView extends Gtk.Box {
     }
 
     _onOverpass(success) {
-        if (!success)
-            return;
-
         this._populate(this.place);
 
         Application.placeStore.addPlace(this.place);
@@ -282,20 +362,20 @@ export class PlaceView extends Gtk.Box {
     _createContent(place) {
         let content = [];
 
-        if (place.isCurrentLocation) {
-            let coordinates = place.location.latitude.toFixed(5)
-                              + ', '
-                              + place.location.longitude.toFixed(5);
+        if (place.isRawCoordinates) {
+            let coordinates = place.coordinatesDescription;
             let accuracyDescription = Utils.getAccuracyDescription(this.place.location.accuracy);
 
             content.push({ label: _("Coordinates"),
-                           icon: 'map-marker-symbolic',
+                           icon: place.isCurrentLocation ?
+                                 'map-marker-symbolic' : 'pin-location-symbolic',
                            info: coordinates });
-
-            content.push({ label: _("Accuracy"),
-                           icon: 'find-location-symbolic',
-                           /* Translators: %s can be "Unknown", "Exact" or "%f km" (or ft/mi/m) */
-                           info: _("Accuracy: %s").format(accuracyDescription) });
+            if (place.isCurrentLocation ||
+                place.location.accuracy !== GeocodeGlib.LOCATION_ACCURACY_UNKNOWN)
+                content.push({ label: _("Accuracy"),
+                               icon: 'find-location-symbolic',
+                               /* Translators: %s can be "Unknown", "Exact" or "%f km" (or ft/mi/m) */
+                               info: _("Accuracy: %s").format(accuracyDescription) });
         }
 
         if (place.website) {
@@ -575,6 +655,16 @@ export class PlaceView extends Gtk.Box {
                 content.push({ label: _("Religion:"), info: religion });
         }
 
+        if (place.isMotorwayJunction && place.ref) {
+            // Translators: This refers to motorway junction exit (freeway exit)
+            const exitRef = _("Exit %s").format(place.ref);
+            const lht = Utils.isLefthandTrafficForCountry(place.countryCode);
+
+            content.push({ icon: lht ? 'arrow2-top-left-symbolic' :
+                                       'arrow2-top-right-symbolic',
+                           info: exitRef });
+        }
+
         return content;
     }
 
@@ -606,7 +696,6 @@ export class PlaceView extends Gtk.Box {
 
             if (icon) {
                 let widget = new Gtk.Image({ icon_name: icon,
-                                             visible: true,
                                              valign: Gtk.Align.START,
                                              halign: Gtk.Align.END });
 
@@ -617,7 +706,6 @@ export class PlaceView extends Gtk.Box {
                 box.append(widget);
             } else if (label) {
                 let widget = new Gtk.Label({ label: label.italics(),
-                                             visible: true,
                                              use_markup: true,
                                              yalign: 0,
                                              halign: Gtk.Align.END });
@@ -634,19 +722,24 @@ export class PlaceView extends Gtk.Box {
             let widget;
 
             if (grid) {
-                widget = new Gtk.Grid({ visible:        true,
-                                        column_spacing: 6 });
+                widget = new Gtk.Grid({ column_spacing: 6,
+                                        row_spacing:    6 });
 
                 for (let i = 0; i < grid.length; i++) {
                     let row = grid[i];
 
                     for (let j = 0; j < row.length; j++) {
-                        let label = new Gtk.Label({ label:   row[j],
-                                                    visible: true,
-                                                    xalign:  0,
-                                                    hexpand: false,
-                                                    wrap: true,
-                                                    halign:  Gtk.Align.FILL });
+                        /* if there's only one component, e.g. in opening
+                         * hours, allow more space
+                         */
+                        const maxChars = row.length === 1 ? 30 : 15;
+                        const label = new Gtk.Label({ label:   row[j],
+                                                      xalign:  0,
+                                                      hexpand: true,
+                                                      wrap: true,
+                                                      max_width_chars: maxChars,
+                                                      halign:  Gtk.Align.FILL,
+                                                      valign:  Gtk.Align.START });
 
                         if (j === 1) {
                             /* set tabular digits for the second column to get
@@ -664,13 +757,13 @@ export class PlaceView extends Gtk.Box {
                 }
             } else {
                 widget = new Gtk.Label({ label: info,
-                                         visible: true,
                                          use_markup: true,
                                          max_width_chars: 30,
                                          wrap: true,
                                          xalign: 0,
                                          hexpand: true,
-                                         halign: Gtk.Align.FILL });
+                                         halign: Gtk.Align.FILL,
+                                         wrap: true });
             }
 
             box.append(widget);
